// GameX
// Vector Class Header
// 
// 2D, 3D, and 4D vectors with elements of any type
// (although the type must be a numeric primitive for full functionality)
// 
// Vector Operations Implemented:
// -- operations that do not modify	the	vector:
//		+, -, *, /,	^			Term-by-term operations (add,	subtract, multiply,	divide,	exponentiate) on vectors (also work	with scalars)
//		+, -					Unary operators	to take	the	equivalent of or negate	a vector
//		==,	!=					Vector equality	and	inequality tests (strict)
//		<, >, <=, >=			Vector length comparison tests
//		Dot						Dot	product	of vector with another vector (sum of multiples	of components)
//		Cross					Cross product of vector	with another vector	(3D	or 2D vectors only)
//		Direction				Gives a	unit-length	vector with	the	same direction as this one
//		DirectionTo				Gives a	unit-length	vector with	a direction of the difference between it and another one
//		Rescaled				Gives a	custom-length vector with the same direction as this one
//		PerpNormal/PerpVec		Gives a vector perpendicular to this one (2D vectors only). PerpVec doesn't normalize it.
//		Reflect/ReflectNormal	Gives this vector reflected off of a line in the direction of another vector (2D vectors only). ReflectNormal returns a normalized result.
//		Length					Length of a	vector
//		LengthSq				Length^2 of	a vector
//		Dist					Distance from vector to	another	vector
//		DistSq					Distance^2 from	vector to another vector
//		Max, Min				Minimum	or maximum component getters
//		Floor, Ceil				Get vector with components rounded up or down to integer boundaries
//		Sum, Avg				Sum	and	mean of	all	components in the vector
//		IsZero, IsNonzero		Determine whether all of the vectors components	are	zero
//		Abs						Gives a	vector that	(term-by-term) is the absolute value of	this one
//		*=						When applied to a Matrix (vec *= mat;), transforms the vector by that matrix
//		[]/(), Get/SetComponent	Lets you access a component of the vector by its index number (0 for x, 1 for y, 2 for z, 3 for w)
// -- operations that modify the vector	being operated on:
//		=, +=, -=, *=, /=, ^=	Term-by-term modifications,	for	example: a *= b	sets a.x = a.x * b.x and a.y = a.y * b.y   (also work with scalars)
//		CrossWith				Sets vector	to be the cross	product	of itself with another vector (3D vectors only)
//		Normalize				Sets vector	to be its Direction	vector and returns the length it used to be
//		MakeAbs					Makes this vector the (term-by-term) absolute value of what it was
//		MakeFloor, MakeCeil		Makes this vector have components rounded up or down to integer boundaries


#ifndef GAMEX_VECTOR_H_INCLUDED
#define GAMEX_VECTOR_H_INCLUDED

#include "gamex-global-def.h"
#include <math.h>
#define T template <typename type>
T struct Vec2; T struct Vec3; T struct Vec4;
//template <int rows, int cols, typename type> class Matrix;
template <typename type, int rows, int cols> struct Matrix;
typedef Matrix<float,4,4> Mat44f; // matrix of 4x4 floats

// define some of the most common vector types
// (you can, of course, define other vector types similarly)
typedef Vec2<float>         Vec2f;  // vector of 2 floats
typedef Vec2<signed short>  Vec2i;  // vector of 2 short ints
typedef Vec3<float>         Vec3f;  // vector of 3 floats
typedef Vec3<signed short>  Vec3i;  // vector of 3 short ints
typedef Vec3<double>        Vec3fd; // vector of 3 double-precision floats
typedef Vec4<float>         Vec4f;  // vector of 4 floats
typedef Vec4<signed short>  Vec4i;  // vector of 4 short ints
typedef Vec4<unsigned char> Vec4c;  // vector of 4 bytes

// alternate naming scheme for (some degree of) backward compatibility
typedef Vec2f  Vector2DF;
typedef Vec2i  Vector2DI;
typedef Vec3f  Vector3DF;
typedef Vec3i  Vector3DI;
typedef Vec3fd Vector3DD;
typedef Vec4f  Vector4DF;
typedef Vec4i  Vector4DI;
typedef Vec4c  Vector4DC;

// forward declaration, used by their counterparts in LineSeg2
T struct LineSeg2;
T bool testIntersection(const LineSeg2<type> & line1, const LineSeg2<type> & line2, Vec2f & outputCollisionPoint);
T bool testIntersectionFast(const LineSeg2<type> & line2, Vec2f & outputCollisionPoint);
T void setupTestIntersectionFast(const LineSeg2<type> & line1);

// a line segment in 2D space
T struct LineSeg2
{
	Vec2<type> p1; // start
	Vec2<type> p2; // end

	// constructors
	LineSeg2<type> (Vec2<type> np1, Vec2<type> np2) { p1 = np1; p2 = np2; }
	LineSeg2<type> () {}

	// cast to vector gives the difference between endpoints
	inline operator Vec2<type> () const { return p2 - p1; }

	// general-purpose test of the intersection of two line segments
	inline bool testIntersection(const LineSeg2<type> & line2, Vec2f & outputCollisionPoint) const { return ::testIntersection(*this, line2, outputCollisionPoint); }

	// faster intersection test for when you are testing 1 line against many other lines
	// (in tests I've run with many lines and optimizations on, it's about 17% faster than testIntersection)
	inline bool testIntersectionFast(const LineSeg2<type> & line2, Vec2f & outputCollisionPoint) const;

	// must be called before testIntersectionFast to tell it which line is being tested against all the others
	inline void setupTestIntersectionFast() const { return ::setupTestIntersectionFast(*this); }
};
typedef LineSeg2<float> LineSeg2f; // a line segment in 2D space of floats


#pragma warning(push)

// disable "too many parameters" warning because that's intentional to get the effect of variable number of params to a macro
#pragma warning(disable: 4002)

// disable "possible loss of data" warnings since int times float can happen a lot naturally depending on which template types are used
#pragma warning(disable: 4244)

#pragma pack(push)
#pragma pack(1)

// crazy helper macros
// Note:  OP2 and 3 defined as variadic to mimic VC's behavior of ignoring
//        extra arguments to macros.
#define APPLY_OP(op, toObj, field) { field op (toObj).field; } // this->field (operator) obj.field;
#define APPLY_OP2(op, toObj, field1, field2, ...) { APPLY_OP(op, toObj, field1); APPLY_OP(op, toObj, field2); }
#define APPLY_OP3(op, toObj, field1, field2, field3, ...) { APPLY_OP(op, toObj, field1); APPLY_OP2(op, toObj, field2, field3); }
#define APPLY_OP4(op, toObj, field1, field2, field3, field4) { APPLY_OP(op, toObj, field1); APPLY_OP3(op, toObj, field2, field3, field4); }
#define APPLY_OP_PRIM(op, prim, field) { field op (prim); } // this->field (operator) primitive;
#define APPLY_OP2_PRIM(op, prim, field1, field2, ...) { APPLY_OP_PRIM(op, prim, field1); APPLY_OP_PRIM(op, prim, field2); }
#define APPLY_OP3_PRIM(op, prim, field1, field2, field3, ...) { APPLY_OP_PRIM(op, prim, field1); APPLY_OP2_PRIM(op, prim, field2, field3); }
#define APPLY_OP4_PRIM(op, prim, field1, field2, field3, field4) { APPLY_OP_PRIM(op, prim, field1); APPLY_OP3_PRIM(op, prim, field2, field3, field4); }
#define T2 template <typename type2>
#define T23 template <typename type2, typename type3>
#define T234 template <typename type2, typename type3, typename type4>
#define T2345 template <typename type2, typename type3, typename type4, typename type5>
#define PR template <typename prim>
#define AN template <typename anything>
#define P_T2 template <typename prim, typename type2>
#define T_T2 template <typename type, typename type2>
#define SQF(x) (float)((x)*(x)) // squared float macro
#define ABS(x) ((x) < 0 ? (-x) : (x)) // absolute value macro
#define VecT(n) Vec##n<type> // vector of type type with n dimensions
#define VecT2(n) Vec##n<type2> // vector of type type2 with n dimensions

// operations and functions that all of the following vector types have
// (not gained through inheritance due to syntax complications and speed concerns)
// this also qualifies as a crazy macro
#define DEFINE_SHARED_VEC_OPS(n)\
T2	inline void operator-= (const VecT2(n)& op) { APPLY_OP##n( -=, op, x,y,z,w ); } \
T2	inline void operator+= (const VecT2(n)& op) { APPLY_OP##n( +=, op, x,y,z,w ); } \
T2	inline void operator*= (const VecT2(n)& op) { APPLY_OP##n( *=, op, x,y,z,w ); } \
T2	inline void operator/= (const VecT2(n)& op) { APPLY_OP##n( /=, op, x,y,z,w ); } \
PR	inline void operator = (const prim& op) { APPLY_OP##n##_PRIM(  =, op, x,y,z,w ); } \
PR	inline void operator-= (const prim& op) { APPLY_OP##n##_PRIM( -=, op, x,y,z,w ); } \
PR	inline void operator+= (const prim& op) { APPLY_OP##n##_PRIM( +=, op, x,y,z,w ); } \
PR	inline void operator*= (const prim& op) { APPLY_OP##n##_PRIM( *=, op, x,y,z,w ); } \
PR	inline void operator/= (const prim& op) { APPLY_OP##n##_PRIM( /=, op, x,y,z,w ); } \
T2	inline bool operator<  (const VecT2(n)& op) const { return LengthSq() < op.LengthSq(); } \
T2	inline bool operator>  (const VecT2(n)& op) const { return LengthSq() > op.LengthSq(); } \
T2	inline bool operator<= (const VecT2(n)& op) const { return LengthSq() <= op.LengthSq(); } \
T2	inline bool operator>= (const VecT2(n)& op) const { return LengthSq() >= op.LengthSq(); } \
AN	inline VecT(n) operator- (const anything& op) const { VecT(n) v (*this); v -= op; return v; } \
AN	inline VecT(n) operator+ (const anything& op) const { VecT(n) v (*this); v += op; return v; } \
AN	inline VecT(n) operator* (const anything& op) const { VecT(n) v (*this); v *= op; return v; } \
AN	inline VecT(n) operator/ (const anything& op) const { VecT(n) v (*this); v /= op; return v; } \
AN	inline VecT(n) operator^ (const anything& op) const { VecT(n) v (*this); v ^= op; return v; } \
T2	inline float Dist (const VecT2(n)& v) const {float distsq = DistSq(v); if(distsq>0.0f) return sqrtf(distsq); return 0.0f;} \
T2	inline VecT(n) DirectionTo (const VecT2(n)& v) const { return (v - *this).Direction(); } \
	inline float Length () const {return sqrtf(LengthSq());} \
	inline bool IsZero () const {return !IsNonzero();} \
	inline VecT(n) Abs () const {VecT(n) v (*this); v.MakeAbs(); return v;} \
	inline VecT(n) Floor () const {VecT(n) v (*this); v.MakeFloor(); return v;} \
	inline VecT(n) Ceil () const {VecT(n) v (*this); v.MakeCeil(); return v;} \
	inline float Avg () const {return (float)Sum() * (1.0f / (float)n);} \
	inline float Normalize () {if(IsNonzero()) {const float len = Length(); *this *= 1.0f/len; return len;} return 0.0f;} \
	inline float Normalize (type normalizedDist) {if(IsNonzero()) {const float len = Length(); *this *= (float)normalizedDist / len; return len;} return 0.0f;} \
	inline VecT(n) Direction () const {if(IsNonzero()) {return *this * 1.0f/Length();} return VecT(n)(0);} \
	inline VecT(n) Rescaled (type normalizedDist) const {if(IsNonzero()) {return *this * ((float)normalizedDist / Length());} return VecT(n)(0);} \
	/*inline operator bool () const {return IsNonzero();}*/ /* cast to bool */ \
	inline const type& operator () (int i) const {return array[i];} \
	inline const type& operator [] (int i) const {return array[i];} \
	inline type& operator () (int i) {return array[i];} \
	inline type& operator [] (int i) {return array[i];} \
	inline void SetComponent (int i, const type& val) {array[i] = val;} \
T2	inline void Set (const VecT2(n)& op) {*this = op;}

// common operations that must be outside of the class definition
#define DEFINE_SHARED_VEC_POST_OPS(n)\
T_T2	inline VecT(n) operator * (const VecT(n)& v1, const VecT2(n)& v2) {return v1.operator *(v2);} /* disambiguates vec*vec from prim*vec */ \
T_T2	inline VecT(n) operator / (const VecT(n)& v1, const VecT2(n)& v2) {return v1.operator /(v2);} /* disambiguates vec/vec from prim/vec */ \
P_T2	inline VecT2(n) operator * (const prim op, const VecT2(n)& v) {return v * op;} \
P_T2	inline VecT2(n) operator / (const prim op, const VecT2(n)& v) {return VecT2(n)(op) / v;} \
T2	inline VecT2(n) operator - (const VecT2(n)& v) {return VecT2(n)(0) - v;} \
T2	inline const VecT2(n)&operator + (const VecT2(n)& v) {return v;} \
T2	inline VecT2(n)&operator + (VecT2(n)& v) {return v;}



// vector with 2 components of any type
T struct Vec2
{
	// you can access the values of a Vec2 by name (like vector.x and vector.y)
	// or by index (like vector[0] and vector[1]),
	// and you can request the underlying array with vector.array
	union
	{
		struct { type x, y; };
		type array [2];
	};

T23	inline void Set (const type2& nx, const type3& ny) {x = (type)nx; y = (type)ny;}
T23	inline Vec2<type> (const type2& nx, const type3& ny) {Set(nx,ny);}
T2	inline Vec2<type> (const Vec2<type2>& v) {*this = v;}
T2	inline Vec2<type> (const type2* n) {x = (type)n[0]; y = (type)n[1];}
	explicit inline Vec2<type> (const type& n) {x = y = (type)n;}
	inline Vec2<type> () {/*x = 0; y = 0;*/}

T2	inline void operator = (const Vec4<type2>& op) { APPLY_OP2(  =, op, x, y ); }
T2	inline void operator = (const Vec3<type2>& op) { APPLY_OP2(  =, op, x, y ); }
T2	inline void operator = (const Vec2<type2>& op) { APPLY_OP2(  =, op, x, y ); }
T2	inline bool operator== (const Vec2<type2>& op) const { return ((x == op.x) && (y == op.y)); }
T2	inline bool operator!= (const Vec2<type2>& op) const { return ((x != op.x) || (y != op.y)); }
T2	inline void operator^= (const Vec2<type2>& op) { x = powf(x, op.x); y = powf(y, op.y); }
PR	inline void operator^= (const prim& op) { x = powf(x, op); y = powf(y, op); }

T2	inline type Dot (const Vec2<type2>& op) const { return x*op.x + y*op.y; }
T2	inline type Cross (const Vec2<type2>& op) const { return y*op.x - x*op.y; }
T2	inline void CrossWith (const Vec2<type2>& v) {assert(!"Vec2::CrossWith() is not implemented, although Vec2::Cross() is.");}
	inline void MakeAbs () {x = ABS(x); y = ABS(y);}
	inline void MakeFloor () {x = floorf(x); y = floorf(y);}
	inline void MakeCeil () {x = ceilf(x); y = ceilf(y);}
	inline Vec2<type> PerpVec () const { return Vec2<type>(-y, x); }
	inline Vec2<type> PerpNormal () const { Vec2<type> v(-y, x); v.Normalize(); return v; }
	inline Vec2<type> Reflect (const Vec2<type> & op) const { Vec2<type> v = op.PerpNormal(); v *= (-2 * Dot(v)); v += *this; return v; }
	inline Vec2<type> ReflectNormal (const Vec2<type> & op) const { Vec2<type> v = op.PerpVec(); v *= (-2 * Dot(v)); v += *this; v.Normalize(); return v; }

	inline bool IsNonzero () const {return x||y;}
	inline float LengthSq () const {return SQF(x) + SQF(y);}
T2	inline float DistSq (const Vec2<type2>& v) const {return SQF(x - v.x) + SQF(y - v.y);}
	inline type Max () const {return x > y ? x : y;}
	inline type Min () const {return x < y ? x : y;}
	inline type Sum () const {return x + y;}

T23	bool IsWithin (const Vec2<type2>& minim, const Vec2<type3>& maxim) const;

	float ToAngleRadians () { return atan2f(x,y); }
	inline float ToAngleRadiansCW () { return -ToAngleRadians(); }
	float ToAngleDegrees () { return atan2f(x,y) * RADtoDEG; }
	inline float ToAngleDegreesCW () { return -ToAngleDegrees(); }
	static Vec2<type> FromAngleDegrees (float angle, type length) { return Vec2<type>(length * cosf(angle*DEGtoRAD), length * sinf(angle*DEGtoRAD)); }
	inline static Vec2<type> FromAngleDegreesCW (float angle, type length) { return FromAngleDegrees(-angle,length); }
	static Vec2<type> FromAngleRadians (float angle, type length) { return Vec2<type>(length * cosf(angle), length * sinf(angle)); }
	inline static Vec2<type> FromAngleRadiansCW (float angle, type length) { return FromAngleRadians(-angle,length); }
	static const Vec2<type> ZeroVec;
	static const Vec2<type> XVec;
	static const Vec2<type> YVec;

	DEFINE_SHARED_VEC_OPS(2);
	// the rest of this class' functions and operators are in here (see above)
};
DEFINE_SHARED_VEC_POST_OPS(2);



// vector with 3 components of any type
T struct Vec3
{
	// you can access the values of a Vec3 by name (like vector.x and vector.y and vector.z)
	// or by index (like vector[0] and vector[1] and vector[2]),
	// and you can request the underlying array with vector.array
	union
	{
		struct { type x, y, z; };
		type array [3];
	};

T234 inline void Set (const type2& nx, const type3& ny, const type4& nz) {x = (type)nx; y = (type)ny; z = (type)nz;}
T234 inline Vec3<type> (const type2& nx, const type3& ny, const type4& nz) {Set(nx,ny,nz);}
T2	 inline Vec3<type> (const Vec3<type2>& v) {*this = v;}
T2	 inline Vec3<type> (const type2* n) {x = (type)n[0]; y = (type)n[1]; z = (type)n[2];}
	 explicit inline Vec3<type> (const type& n) {x = y = z = (type)n;}
	 inline Vec3<type> () {/*x = 0; y = 0; z = 0;*/}

T2	inline void operator = (const Vec4<type2>& op) { APPLY_OP3(  =, op, x, y, z ); }
T2	inline void operator = (const Vec3<type2>& op) { APPLY_OP3(  =, op, x, y, z ); }
T2	inline void operator = (const Vec2<type2>& op) { APPLY_OP2(  =, op, x, y ); }
T2	inline bool operator== (const Vec3<type2>& op) const { return ((x == op.x) && (y == op.y) && (z == op.z)); }
T2	inline bool operator!= (const Vec3<type2>& op) const { return ((x != op.x) || (y != op.y) || (z != op.z)); }
T2	inline void operator^= (const Vec3<type2>& op) { x = powf(x, op.x); y = powf(y, op.y); z = powf(z, op.z); }
PR	inline void operator^= (const prim& op) { x = powf(x, op); y = powf(y, op); z = powf(z, op); }

T2	inline type Dot (const Vec3<type2>& op) const {return x*op.x + y*op.y + z*op.z;}
T2	inline Vec3<type> Cross (const Vec3<type2>& op) const {Vec3<type> v (*this); v.CrossWith(op); return v;}
T2	inline void CrossWith (const Vec3<type2>& v) {float ax = x, ay = y; x = (ay * v.z - z * v.y); y = (-ax * v.z + z * v.x); z = (ax * v.y - ay * v.x);}
	inline void MakeAbs () {x = ABS(x); y = ABS(y); z = ABS(z);}
	inline void MakeFloor () {x = floorf(x); y = floorf(y); z = floorf(z);}
	inline void MakeCeil () {x = ceilf(x); y = ceilf(y); z = ceilf(z);}

	inline bool IsNonzero () const {return x||y||z;}
	inline float LengthSq () const {return SQF(x) + SQF(y) + SQF(z);}
T2	inline float DistSq (const Vec3<type2>& v) const {return SQF(x - v.x) + SQF(y - v.y) + SQF(z - v.z);}
	inline type Max () const {return x > y ? (x > z ? x : z) : (y > z ? y : z);}
	inline type Min () const {return x < y ? (x < z ? x : z) : (y < z ? y : z);}
	inline type Sum () const {return x + y + z;}

	void operator*= (Mat44f& op);
	inline void TransformAsPointBy (Mat44f& op) { *this *= op; }
	void TransformAsVectorBy (Mat44f& op);

T23	bool IsWithin (const Vec3<type2>& minim, const Vec3<type3>& maxim) const;

	static const Vec3<type> ZeroVec;
	static const Vec3<type> XVec;
	static const Vec3<type> YVec;
	static const Vec3<type> ZVec;

	DEFINE_SHARED_VEC_OPS(3);
	// the rest of this class' functions and operators are in here (see above)
};
DEFINE_SHARED_VEC_POST_OPS(3);


// vector with 4 components of any type
T struct Vec4
{
	// you can access the values of a Vec4 by name (like vector.x and vector.y and vector.y and vector.z)
	// or by index (like vector[0] and vector[1] and vector[2] and vector[3]),
	// and you can request the underlying array with vector.array
	union
	{
		struct { type x, y, z, w; };
		type array [4];
	};

T2345 inline void Set (const type2& nx, const type3& ny, const type4& nz, const type5& nw) {x = (type)nx; y = (type)ny; z = (type)nz; w = (type)nw;}
T2345 inline Vec4<type> (const type2& nx, const type3& ny, const type4& nz, const type5& nw) {Set(nx,ny,nz,nw);}
T2	  inline Vec4<type> (const Vec4<type2>& v) {*this = v;}
T2	  inline Vec4<type> (const type2* n) {x = (type)n[0]; y = (type)n[1]; z = (type)n[2]; w = (type)n[3];}
	  explicit inline Vec4<type> (const type& n) {x = y = z = w = (type)n;}
	  inline Vec4<type> () {/*x = 0; y = 0; z = 0; w = 0;*/}

T2	inline void operator = (const Vec4<type2>& op) { APPLY_OP4(  =, op, x, y, z, w ); }
T2	inline void operator = (const Vec3<type2>& op) { APPLY_OP3(  =, op, x, y, z ); }
T2	inline void operator = (const Vec2<type2>& op) { APPLY_OP2(  =, op, x, y ); }
T2	inline bool operator== (const Vec4<type2>& op) const { return ((x == op.x) && (y == op.y) && (z == op.z) && (w == op.w)); }
T2	inline bool operator!= (const Vec4<type2>& op) const { return ((x != op.x) || (y != op.y) || (z != op.z) || (w != op.w)); }
T2	inline void operator^= (const Vec4<type2>& op) { x = powf(x, op.x); y = powf(y, op.y); z = powf(z, op.z); w = powf(w, op.w); }
PR	inline void operator^= (const prim& op) { x = powf(x, op); y = powf(y, op); z = powf(z, op); w = powf(w, op); }

T2	inline type Dot (const Vec4<type2>& op) const {return x*op.x + y*op.y + z*op.z + w*op.w;}
T2	inline Vec4<type> Cross (const Vec4<type2>& op) const {Vec4<type> v (*this); v.CrossWith(op); return v;}
T2	inline void CrossWith (const Vec4<type2>& v) {assert(!"Vec4::CrossWith() and Vec4::Cross() are not implemented.");}
	inline void MakeAbs () {x = ABS(x); y = ABS(y); z = ABS(z); w = ABS(w);}
	inline void MakeFloor () {x = floorf(x); y = floorf(y); z = floorf(z); w = floorf(w);}
	inline void MakeCeil () {x = ceilf(x); y = ceilf(y); z = ceilf(z); w = ceilf(w);}

	inline bool IsNonzero () const {return x||y||z||w;}
	inline float LengthSq () const {return SQF(x) + SQF(y) + SQF(z) + SQF(w);}
T2	inline float DistSq (const Vec4<type2>& v) const {return SQF(x - v.x) + SQF(y - v.y) + SQF(z - v.z) + SQF(w - v.w);}
	inline type Max () const {return x > y ? (x > z ? (x > w ? x : w) : (z > w ? z : w)) : (y > z ? (y > w ? y : w) : (z > w ? z : w));}
	inline type Min () const {return x < y ? (x < z ? (x < w ? x : w) : (z < w ? z : w)) : (y < z ? (y < w ? y : w) : (z < w ? z : w));}
	inline type Sum () const {return x + y + z + w;}

	void operator*= (Mat44f& op);

T23	bool IsWithin (const Vec4<type2>& minim, const Vec4<type3>& maxim) const;

	static const Vec4<type> ZeroVec;
	static const Vec4<type> XVec;
	static const Vec4<type> YVec;
	static const Vec4<type> ZVec;
	static const Vec4<type> WVec;

	DEFINE_SHARED_VEC_OPS(4);
	// the rest of this class' functions and operators are in here (see above)
};
DEFINE_SHARED_VEC_POST_OPS(4);


// Values

T const Vec2<type> Vec2<type>::ZeroVec = Vec2<type>(0);
T const Vec2<type> Vec2<type>::XVec = Vec2<type>(1,0);
T const Vec2<type> Vec2<type>::YVec = Vec2<type>(0,1);

T const Vec3<type> Vec3<type>::ZeroVec = Vec3<type>(0);
T const Vec3<type> Vec3<type>::XVec = Vec3<type>(1,0,0);
T const Vec3<type> Vec3<type>::YVec = Vec3<type>(0,1,0);
T const Vec3<type> Vec3<type>::ZVec = Vec3<type>(0,0,1);

T const Vec4<type> Vec4<type>::ZeroVec = Vec4<type>(0);
T const Vec4<type> Vec4<type>::XVec = Vec4<type>(1,0,0,0);
T const Vec4<type> Vec4<type>::YVec = Vec4<type>(0,1,0,0);
T const Vec4<type> Vec4<type>::ZVec = Vec4<type>(0,0,1,0);
T const Vec4<type> Vec4<type>::WVec = Vec4<type>(0,0,0,1);


// Global overrides

// bounding box (AABB)

T inline Vec2<type> & Clamp (Vec2<type> & variable, const Vec2<type> & minim, const Vec2<type> & maxim)
{
	if(variable.x < minim.x) variable.x = minim.x;
	else if(variable.x > maxim.x) variable.x = maxim.x;
	if(variable.y < minim.y) variable.y = minim.y;
	else if(variable.y > maxim.y) variable.y = maxim.y;
	return variable;
}
T inline const Vec2<type> & Clamp (const Vec2<type> & variable, const Vec2<type> & minim, const Vec2<type> & maxim)
{
	return Vec2<type>(max(minim.x, min(maxim.x, variable.x)), max(minim.y, min(maxim.y, variable.y)));
}

T inline Vec3<type> & Clamp (Vec3<type> & variable, const Vec3<type> & minim, const Vec3<type> & maxim)
{
	if(variable.x < minim.x) variable.x = minim.x;
	else if(variable.x > maxim.x) variable.x = maxim.x;
	if(variable.y < minim.y) variable.y = minim.y;
	else if(variable.y > maxim.y) variable.y = maxim.y;
	if(variable.z < minim.z) variable.z = minim.z;
	else if(variable.z > maxim.z) variable.z = maxim.z;
	return variable;
}
T inline const Vec3<type> & Clamp (const Vec3<type> & variable, const Vec3<type> & minim, const Vec3<type> & maxim)
{
	return Vec3<type>(max(minim.x, min(maxim.x, variable.x)), max(minim.y, min(maxim.y, variable.y)), max(minim.z, min(maxim.z, variable.z)));
}

T inline Vec4<type> & Clamp (Vec4<type> & variable, const Vec4<type> & minim, const Vec4<type> & maxim)
{
	if(variable.x < minim.x) variable.x = minim.x;
	else if(variable.x > maxim.x) variable.x = maxim.x;
	if(variable.y < minim.y) variable.y = minim.y;
	else if(variable.y > maxim.y) variable.y = maxim.y;
	if(variable.z < minim.z) variable.z = minim.z;
	else if(variable.z > maxim.z) variable.z = maxim.z;
	if(variable.w < minim.w) variable.w = minim.w;
	else if(variable.w > maxim.w) variable.w = maxim.w;
	return variable;
}
T inline const Vec4<type> & Clamp (const Vec4<type> & variable, const Vec4<type> & minim, const Vec4<type> & maxim)
{
	return Vec4<type>(max(minim.x, min(maxim.x, variable.x)), max(minim.y, min(maxim.y, variable.y)), max(minim.z, min(maxim.z, variable.z)), max(minim.w, min(maxim.w, variable.w)));
}

/* // maybe this math is wrong?
T void Vec4<type>::CrossWith (const Vec4<type>& v, const Vec4<type>& v2)
{
	float ax =    x, ay =    y, az =    z, aw =    w;
	float bx =  v.x, by =  v.y, bz =  v.z, bw =  v.w;
	float cx = v2.x, cy = v2.y, cz = v2.z, cw = v2.w;
	float a = (bx * cy) - (by * cx);
	float b = (bx * cz) - (bz * cx);
	float c = (bx * cw) - (bw * cx);
	float d = (by * cz) - (bz * cy);
	float e = (by * cw) - (bw * cy);
	float f = (bz * cw) - (bw * cz);
	x = ((ay * f) - (az * e) + (aw * d));
	y = ((ax * f) + (az * c) - (aw * b));
	z = ((ax * e) - (ay * c) + (aw * a));
	w = (-(ax * d) + (ay * b) - (az * a));
	return *this;
}
*/

T T23 bool Vec2<type>::IsWithin (const Vec2<type2>& minim, const Vec2<type3>& maxim) const
{
	return (x >= minim.x && x <= maxim.x && 
			y >= minim.y && y <= maxim.y);
}
T T23 bool Vec3<type>::IsWithin (const Vec3<type2>& minim, const Vec3<type3>& maxim) const
{
	return (x >= minim.x && x <= maxim.x && 
			y >= minim.y && y <= maxim.y && 
			z >= minim.z && z <= maxim.z);
}
T T23 bool Vec4<type>::IsWithin (const Vec4<type2>& minim, const Vec4<type3>& maxim) const
{
	return (x >= minim.x && x <= maxim.x && 
			y >= minim.y && y <= maxim.y && 
			z >= minim.z && z <= maxim.z && 
			w >= minim.w && w <= maxim.w);
}




// general-purpose test of the intersection of two 2D line segments
T bool testIntersection(const LineSeg2<type> & line1, const LineSeg2<type> & line2, Vec2f & outputCollisionPoint)
{
	const float rise1 = line1.p2.y - line1.p1.y;
	const float run1 = line1.p2.x - line1.p1.x;
	const float rise2 = line2.p2.y - line2.p1.y;
	const float run2 = line2.p2.x - line2.p1.x;

	if(rise1 != 0.0f && run1 != 0.0f)
	{
		const float xDiff = line2.p1.x - line1.p1.x;
		const float yDiff = line2.p1.y - line1.p1.y;
		const float invslope1 = run1 / rise1;
		const float factor2 = (xDiff - yDiff * invslope1) / (rise2 * invslope1 - run2);

		if(factor2 >= 0.0f && factor2 <= 1.0f)
		{
			const float factor1 = (xDiff + factor2 * run2) / run1;
			if(factor1 >= 0.0f && factor1 <= 1.0f)
			{
				outputCollisionPoint.Set(line1.p1.x + factor1 * run1, line1.p1.y + factor1 * rise1);
				return true;
			}
		}
		return false;
	}
	else if(rise2 != 0.0f && run2 != 0.0f)
	{
		// line 1 is totally horizontal or vertical - try again with lines 1 and 2 switched
		return testIntersection(line2, line1, outputCollisionPoint);
	}
	else if(rise2 != 0.0f)
	{
		// line 1 is totally horizontal and line 2 is totally vertical
		if(line2.p1.y >= line1.p1.y && line2.p2.y <= line1.p1.y
		|| line2.p1.y <= line1.p1.y && line2.p2.y >= line1.p1.y)
		{
			outputCollisionPoint.Set(line2.p1.x, line1.p1.y);
			return true;
		}
		return false;
	}
	else
	{
		// line 1 is totally vertical and line 2 is totally horizontal
		if(line1.p1.y >= line2.p1.y && line1.p2.y <= line2.p1.y
		|| line1.p1.y <= line2.p1.y && line1.p2.y >= line2.p1.y)
		{
			outputCollisionPoint.Set(line1.p1.x, line2.p1.y);
			return true;
		}
		return false;
	}
}

static float z_cached_rise1 = 0.0f;
static float z_cached_run1 = 0.0f;
static float z_cached_invrun1 = 0.0f;
static float z_cached_line1_p1_x = 0.0f;
static float z_cached_line1_p1_y = 0.0f;
static float z_cached_invslope1 = 0.0f;

T inline bool LineSeg2<type>::testIntersectionFast(const LineSeg2<type> & line2, Vec2f & outputCollisionPoint) const { assertMsgNoLib(z_cached_line1_p1_x == (float)p1.x && z_cached_line1_p1_y == (float)p1.y, "Must call setupTestIntersectionFast before testIntersectionFast"); return ::testIntersectionFast(line2, outputCollisionPoint); }

// must be called before testIntersectionFast to tell it which line is being tested against all the others
T void setupTestIntersectionFast(const LineSeg2<type> & line1) 
{
	z_cached_line1_p1_x = line1.p1.x;
	z_cached_line1_p1_y = line1.p1.y;
	z_cached_rise1 = line1.p2.y - z_cached_line1_p1_y;
	z_cached_run1 = line1.p2.x - z_cached_line1_p1_x;
	if(z_cached_rise1 == 0.0f) z_cached_rise1 += 0.01f; // hack
	if(z_cached_run1 == 0.0f) z_cached_run1 += 0.01f; // hack
	z_cached_invslope1 = z_cached_run1 / z_cached_rise1;
	z_cached_invrun1 = 1.0f / z_cached_run1;
}

// faster 2D intersection test for when you are testing 1 line against many other lines
// (in tests I've run with many lines and optimizations on, it's about 17% faster than testIntersection)
T bool testIntersectionFast(const LineSeg2<type> & line2, Vec2f & outputCollisionPoint)
{
	assertNoLib(z_cached_rise1 != 0.0f && z_cached_run1 != 0.0f && z_cached_invslope1 == z_cached_run1 / z_cached_rise1 && z_cached_invrun1 == 1.0f / z_cached_run1);

	const float rise2 = line2.p2.y - line2.p1.y;
	const float run2 = line2.p2.x - line2.p1.x;
	const float xDiff = line2.p1.x - z_cached_line1_p1_x;
	const float yDiff = line2.p1.y - z_cached_line1_p1_y;
	const float factor2 = (xDiff - yDiff * z_cached_invslope1) / (rise2 * z_cached_invslope1 - run2);

	if(factor2 >= 0.0f && factor2 <= 1.0f)
	{
		const float factor1 = (xDiff + factor2 * run2) * z_cached_invrun1;
		if(factor1 >= 0.0f && factor1 <= 1.0f)
		{
			outputCollisionPoint.Set(z_cached_line1_p1_x + factor1 * z_cached_run1, z_cached_line1_p1_y + factor1 * z_cached_rise1);
			return true;
		}
	}
	return false;
}




// undefine helper macros
#undef APPLY_OP
#undef APPLY_OP2
#undef APPLY_OP3
#undef APPLY_OP4
#undef APPLY_OP_PRIM
#undef APPLY_OP2_PRIM
#undef APPLY_OP3_PRIM
#undef APPLY_OP4_PRIM
#undef T2
#undef PR
#undef AN
#undef SQF
#undef ABS
#undef DEFINE_SHARED_VEC_OPS
#undef DEFINE_SHARED_VEC_POST_OPS
#undef T
#undef TYPEDEF


#pragma pack(pop)

// re-enable warnings
#pragma warning(pop)

#endif