//
// GameX - High Performance Timer functions
// 
// This software is released under the GameX GNU GPL 
// Open Source License. See the GameX documentation included
// with this source code for terms of modification, 
// distribution and re-release. 

// returns the current time since startup with as much precision as this hardware supports,
// this will probably be in some ultra-precise unit, could be in fractions of a nanosecond
// (use GetClocksPerSec() to find out exactly how many of these units are in a second)
inline long long GetClock()
{
	LARGE_INTEGER time;
	if(QueryPerformanceCounter(&time)) // get the High Performance Timer count
		return (long long)time.QuadPart;
	else // if QueryPerformanceCounter() failed, that means the hardware doesn't support high-resolution timers
		return (long long)clock(); // so use the built-in low-resolution timer, clock(), instead. (only accurate to a 100th of a second)
}

// returns the number of time units (that GetClock() counts in) that are in one second.
// for instance, if this function returns 3579545, that means there are 3,579,545 clock ticks per second.
// note: you can assume the return value is a constant, although not across different computers.
inline long long GetClocksPerSec()
{
	LARGE_INTEGER frequency;
	if(QueryPerformanceFrequency(&frequency)) // get the High Performance Timer frequency in Hz
		return (long long)frequency.QuadPart;
	else // if QueryPerformanceFrequency() failed, that means the hardware doesn't support high-resolution timers
		return (long long)CLOCKS_PER_SEC; // so use the low-resolution timer's "per second" value
}

// sleeps for at least the given number of time units,
// at most a bit over that amount,
// and increments GameX's idle counter accordingly.
// will only Sleep() in units of sleep_step milliseconds,
// so higher sleep_step means more accuracy but less actual sleeping  
// sleep_step should be an integer from 1 to 10 inclusive,
// clocks is the amount of time to sleep/delay in terms of GetClocksPerSec(), NOT milliseconds
inline void AccurateSleep(long long clocks, int sleep_step = 1)
{
	if(clocks <= 0) return; // wouldn't want to freeze upon sleeping a negative amount, also return quickly on 0 sleep
	long long sleeptimer = GetClock();
	int sleepAmount = (int)((clocks+1) / (GetClocksPerSec() / CLOCKS_PER_SEC));
	if(sleep_step != 1)
		sleepAmount -= (sleepAmount % sleep_step);
	if(sleepAmount > 0)
	{
		Sleep(sleepAmount); // sleep for the amount of time that fits into sleep_step millisecond units
		GameX.z_sleep_record += GetClock() - sleeptimer;
	}
	while(GetClock() - sleeptimer < clocks) ; // loop tightly for any time leftover
}
